/*
 *  This Code Was Created By Jeff Molofee 2000
 *  A HUGE Thanks To Fredric Echols For Cleaning Up
 *  And Optimizing The Base Code, Making It More Flexible!
 *  If You've Found This Code Useful, Please Let Me Know.
 *  Visit My Site At nehe.gamedev.net
 */

#include <windows.h>  // Header File For Windows
#include <time.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <math.h>
#include <gl\gl.h>   // Header File For The OpenGL32 Library
#include <gl\glu.h>   // Header File For The GLu32 Library
#include <gl\glaux.h>  // Header File For The Glaux Library

#include "glext.h"
#include "wglext.h"

#include "projtype.h"
#include "projfile.h"
#include "global.h"

unsigned long *idx;
float* dt; // vertices
float* dtorg; // copy of original vertices
float* nv; // normals
unsigned long dtcount;

double angle, angle2;

GLfloat LightAmbient[]=  { 0.1f, 0.1f, 0.1f, 1.0f };
GLfloat LightDiffuse[]=  { 0.2f, 0.2f, 0.2f, 1.0f };
GLfloat LightSpecular[]=  { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat LightPosition[]= { -50.0f, 50.0f, 10.0f, 1.0f };


GLvoid ReSizeGLScene(GLsizei width, GLsizei height)  // Resize And Initialize The GL Window
{
  if (height==0)          // Prevent A Divide By Zero By
    {
      height=1;          // Making Height Equal One
    }

  glViewport(0,0,width,height);      // Reset The Current Viewport

  glMatrixMode(GL_PROJECTION);      // Select The Projection Matrix
  glLoadIdentity();         // Reset The Projection Matrix

  // Calculate The Aspect Ratio Of The Window
  gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,0.1f,100.0f);

  glMatrixMode(GL_MODELVIEW);       // Select The Modelview Matrix
  glLoadIdentity();         // Reset The Modelview Matrix
}

int InitGL(GLvoid)          // All Setup For OpenGL Goes Here
{
  glShadeModel(GL_SMOOTH );       // Enable Smooth Shading
  glClearColor(0.0f, 0.0f, 0.0f, 0.1f);    // Black Background
  glClearDepth(1.0f);         // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);       // Enables Depth Testing
  glDepthFunc(GL_LEQUAL );        // The Type Of Depth Testing To Do
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // Really Nice Perspective Calculations

  glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0);
  glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0);

  glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);  // Setup The Ambient Light
  glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);  // Setup The Diffuse Light
  glLightfv(GL_LIGHT1, GL_DIFFUSE, LightSpecular);  // Setup The Specular Light
  glLightfv(GL_LIGHT1, GL_SPECULAR,LightPosition); // Position The Light
  glEnable(GL_LIGHT1);        // Enable Light One
  glEnable(GL_LIGHTING);        // Enable Lighting

  // glEnable(GL_NORMALIZE); // better do this yourself

  // glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);


  //glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
  //glBlendFunc(GL_SRC_ALPHA, GL_DST_ALPHA); 

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);

#ifdef USE_GL_NV_VERTEX_ARRAY_RANGE
  glEnableClientState(GL_VERTEX_ARRAY_RANGE_NV);
  glVertexArrayRangeNV = (PFNGLVERTEXARRAYRANGENVPROC)
    wglGetProcAddress("glVertexArrayRangeNV");
  wglAllocateMemoryNV = (PFNWGLALLOCATEMEMORYNVPROC)
    wglGetProcAddress("wglAllocateMemoryNV");
  wglFreeMemoryNV     = (PFNWGLFREEMEMORYNVPROC)
    wglGetProcAddress("wglFreeMemoryNV");
  glFlushVertexArrayRangeNV = (PFNGLFLUSHVERTEXARRAYRANGENVPROC)
    wglGetProcAddress("glFlushVertexArrayRangeNV");
#endif

  return TRUE;          // Initialization Went OK
}

void calc_normals()
{
  float x,y,z, x1,y1,z1, x2,y2,z2, a;
  unsigned long v, strip, square, v1, v2, v3, v4, t, swap;

  memset(nv, 0, 3*dtcount*sizeof(float));
  for(strip=0;strip<XDOTS-1;strip++) 
    for(square=0;square<YDOTS-1;square++) {

      v1 = (strip  ) * YDOTS + square  ;
      v2 = (strip+1) * YDOTS + square  ;
      v3 = (strip  ) * YDOTS + square+1;
      v4 = (strip+1) * YDOTS + square+1;
      // there are two triangles in the square: v2v1v3 and v3v4v2
	
      for(t=0;t<2;t++) {
	if(t==1) {v1=v4;swap=v2;v2=v3;v3=swap;} // rename for second triangle
	// take two sides and do a cross product
	x1 = dt[ 3*v2  ] - dt[ 3*v1  ];
	y1 = dt[ 3*v2+1] - dt[ 3*v1+1];
	z1 = dt[ 3*v2+2] - dt[ 3*v1+2];
	x2 = dt[ 3*v3  ] - dt[ 3*v1  ];
	y2 = dt[ 3*v3+1] - dt[ 3*v1+1];
	z2 = dt[ 3*v3+2] - dt[ 3*v1+2];
	// calculate normal for this triangle
	x = y1 * z2 - y2 * z1;
	y = z1 * x2 - z2 * x1;
	z = x1 * y2 - x2 * y1;
#if 0
	// Let all triangles have the same influence on their vertices.
	a = (float)sqrt(x*x+y*y+z*z);
	x /= a;
	y /= a;
	z /= a;
#else
	// Let big triangles "bend" the normals more than small triangles.
#endif
	// add normal to all vertices of the triangle
	nv[ 3*v1 + 0 ] += x; nv[ 3*v1 + 1 ] += y; nv[ 3*v1 + 2 ] += z;
	nv[ 3*v2 + 0 ] += x; nv[ 3*v2 + 1 ] += y; nv[ 3*v2 + 2 ] += z;
	nv[ 3*v3 + 0 ] += x; nv[ 3*v3 + 1 ] += y; nv[ 3*v3 + 2 ] += z;
      }
    }
      
  // normalize summed up "normal" vectors
  for(v=0;v<dtcount;v++) {
    a = (float)(1.0 / sqrt(   nv[3*v  ]*nv[3*v  ] 
			      + nv[3*v+1]*nv[3*v+1] 
			      + nv[3*v+2]*nv[3*v+2]));
    nv[3*v] *= a; nv[3*v+1] *= a; nv[3*v+2] *= a;
  }

} // calc_normals

int ReadSceneData(void)
{
  char *errmsg;
  char msg[256];

  // read number of dots and dots themself
  if (0!=(errmsg=ReadDots(dtcount,dt))) {
    MessageBox(NULL,"Failed to read dots. ABORT","ReadSceneData",MB_OK | MB_ICONINFORMATION);
    return 0;
  }

  assert(dtcount == XDOTS * YDOTS);

  // make a copy of the dots
  dtorg = new float[3*dtcount];
  if(0 == dtorg) {
    MessageBox(NULL,"Failed to allocate space for dot coopies. ABORT","ReadSceneData",MB_OK | MB_ICONINFORMATION);
    return 0;
  }
  memcpy(dtorg, dt, 3*dtcount*sizeof(float)); 

  // calculate normals
#ifdef USE_GL_NV_VERTEX_ARRAY_RANGE
  nv = (float*) wglAllocateMemoryNV( 3 * dtcount * sizeof(float), 0, 0.1f, 1);
  if(0 == nv) {
    MessageBox(NULL,"Failed to allocate space for normals via wglAllocateMemoryNV.\nUsing normal malloc/new instead.","ReadSceneData",MB_OK | MB_ICONINFORMATION);
    nv = new float[3*dtcount];
  }
#else
    nv = new float[3*dtcount];
#endif

  if(0 == nv) {
    MessageBox(NULL,"Failed to allocate space for normals. ABORT","ReadSceneData",MB_OK | MB_ICONINFORMATION);
    return 0;
  }

  calc_normals();
 
  if(0 && !fullscreen) {
    sprintf(msg, "Read %lu points.", dtcount);
    MessageBox(NULL, msg, "ReadSceneData", MB_OK | MB_ICONINFORMATION);
  }

  return 1;
} /* ReadSceneData */

int DrawGLScene(GLvoid)         // Here's Where We Do All The Drawing
{
  static int firsttime = 1;
  static GLuint numdlist;
  /* unsigned long i,j,k; */
  float material[4], param, factor;
  int strip, square, vnum;
  static unsigned short vidx[(XDOTS-1)*YDOTS*2];
  static double fps = 1;
  char msg [512];

  /* glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); */ // Clear Screen And Depth Buffer
  glLoadIdentity();         // Reset The Current Modelview Matrix
  glTranslated( 0.0, 1.0, -50.0);
  glRotated( angle,  0, 0, 1);
  glRotated( angle2, 1, 0, 0);


  angle = angle + 5.0 * 0.1;
  if(angle>360) angle -= 360.0;

  angle2 = angle2 + 17.0 * 0.1;
  if(angle2>360) angle2 -= 360.0;

  if(firsttime) {
    start = time(0);
  }

  if(fps && framecount > 100) {
    factor = (float)(cos(0.25 * (double)(framecount-100) * 2.0 * M_PI / fps));
    for(vnum=0; vnum<dtcount; vnum++) {
      dt[3*vnum+2] = dtorg[3*vnum+2] * factor;
    }
    calc_normals();
  } else {
    if(framecount == 0) {
      memcpy(dt, dtorg, 3*dtcount*sizeof(float));
    }
    calc_normals();
  }
  glVertexPointer(3, GL_FLOAT, 0, dt);
  glNormalPointer(GL_FLOAT, 0, nv);


#ifdef USE_GL_NV_VERTEX_ARRAY_RANGE
  if(1 || firsttime /* whence? */) {
    glVertexArrayRangeNV(dtcount* 3 * sizeof(dt[0]), dt);
    glVertexArrayRangeNV(dtcount* 3 * sizeof(nv[0]), nv);
  }
#endif



  {material[0]=0.5f;material[1]=0.5f;material[2]=0.9f;material[3]=0.5f;}
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
  {material[0]=0.0f;material[1]=0.0f;material[2]=0.5f;material[3]=0.5f;}
  glMaterialfv(GL_FRONT, GL_SPECULAR, material);
  param = 64; glMaterialfv(GL_FRONT, GL_SHININESS, &param);
  
  {material[0]=0.9f;material[1]=0.5f;material[2]=0.5f;material[3]=0.5f;}
  glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, material);
  {material[0]=-0.5f;material[1]=0.0f;material[2]=0.0f;material[3]=0.5f;}
  glMaterialfv(GL_BACK, GL_SPECULAR, material);
  param = 64; glMaterialfv(GL_BACK, GL_SHININESS, &param);

  if(firsttime) {
    vnum = 0;
    for(strip=0;strip<XDOTS-1;strip+=1) {
      glBegin(GL_TRIANGLE_STRIP);
      for(square=0;square<YDOTS;square++) {
#if 0
	glNormal3fv( nv + 3*( (strip    ) * YDOTS + square) );
	glVertex3fv( dt + 3*( (strip    ) * YDOTS + square) );
	
	glNormal3fv( nv + 3*( (strip + 1) * YDOTS + square) );
	glVertex3fv( dt + 3*( (strip + 1) * YDOTS + square) );
#else	
	glArrayElement( (strip    ) * YDOTS + square);
	glArrayElement( (strip + 1) * YDOTS + square);
	
	vidx[vnum++] = (strip    ) * YDOTS + square;
	vidx[vnum++] = (strip + 1) * YDOTS + square;
#endif
      }
      glEnd(); // GL_TRIANGLE_STRIP
    }
  } else {
    /* array is already initialized: */
    for(strip=0;strip<XDOTS-1;strip+=1) {
      glDrawElements(GL_TRIANGLE_STRIP, 2*YDOTS, GL_UNSIGNED_SHORT, 
		     vidx + 2*YDOTS*strip);
    }
  }

  // if( GL_NO_ERROR != glGetError()) MessageBox(NULL,"Error in glDrawElements","What a mess!",MB_OK | MB_ICONINFORMATION);

  framecount++;

  stop = time(0);
  if(stop-start) fps =  ((fps * framecount) + (framecount / (stop-start)))
		   / (framecount+1);
  if(0 == framecount % 5) {
    sprintf(msg, "Sombrero - ESC quits and F1 toggles fullscreen. %lu frames, %lu s, %0.2f fps", 
	    framecount, stop-start, fps);
    SetWindowText(hWnd, msg);
  }


  firsttime=0;
#ifdef USE_GL_NV_VERTEX_ARRAY_RANGE
  //glFlushVertexArrayRangeNV(); // should we ?
#endif
  return TRUE;          // Keep Going
}

GLvoid KillGLWindow(GLvoid)        // Properly Kill The Window
{
  if (fullscreen)          // Are We In Fullscreen Mode?
    {
      ChangeDisplaySettings(NULL,0);     // If So Switch Back To The Desktop
      ShowCursor(TRUE);        // Show Mouse Pointer
    }

  if (hRC)           // Do We Have A Rendering Context?
    {
      if (!wglMakeCurrent(NULL,NULL))     // Are We Able To Release The DC And RC Contexts?
	{
	  MessageBox(NULL,"Release Of DC And RC Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
	}

      if (!wglDeleteContext(hRC))      // Are We Able To Delete The RC?
	{
	  MessageBox(NULL,"Release Rendering Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
	}
      hRC=NULL;          // Set RC To NULL
    }

  if (hDC && !ReleaseDC(hWnd,hDC))     // Are We Able To Release The DC
    {
      MessageBox(NULL,"Release Device Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
      hDC=NULL;          // Set DC To NULL
    }

  if (hWnd && !DestroyWindow(hWnd))     // Are We Able To Destroy The Window?
    {
      MessageBox(NULL,"Could Not Release hWnd.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
      hWnd=NULL;          // Set hWnd To NULL
    }

  if (!UnregisterClass("OpenGL",hInstance))   // Are We Able To Unregister Class
    {
      MessageBox(NULL,"Could Not Unregister Class.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
      hInstance=NULL;         // Set hInstance To NULL
    }

#ifdef USE_GL_NV_VERTEX_ARRAY_RANGE  
  wglFreeMemoryNV(dt);
  wglFreeMemoryNV(nv);
#else
  delete [] dt;
  delete [] nv;
#endif
}




